# Copyright (c) 2021 Al Sweigart
# Copyright (c) 2023 Andy Brody
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Adapted from pwinput, by Al Sweigart

A cross-platform Python module that displays **** for password input.
"""


import sys
from typing import List


def pwinput(prompt: str = "Password: ", mask: str = "•") -> str:
    """
    Like getpass.getpass(), but echo the mask character to stdout with each
    keystroke.
    """

    if not isinstance(prompt, str):
        raise TypeError(
            "prompt argument must be a str, not %s" % (type(prompt).__name__)
        )
    if not isinstance(mask, str):
        raise TypeError("mask argument must be a str, not %s" % (type(prompt).__name__))
    if len(mask) > 1:
        raise ValueError("mask argument must be a zero- or one-character str")

    if mask == "" or sys.stdin is not sys.__stdin__ or not sys.stdin.isatty():
        # Just use getpass if a mask is not needed.
        # Note that getpass will attempt to read/write directly from /dev/tty,
        # so if stdin is a pipe, getpass may still read from the terminal.
        import getpass

        return getpass.getpass(prompt)

    enteredPassword: List[str] = []

    sys.stdout.write(prompt)
    sys.stdout.flush()

    while True:
        key = ord(term_getchar(strip_escapes=True))

        if key == 13 or key == 4:
            # enter key or ^D pressed
            sys.stdout.write("\n")
            return "".join(enteredPassword)
        elif key in (8, 127):  # Backspace/Del key erases previous output.
            if len(enteredPassword) > 0:
                # Erase previous character
                # Print \b to move cursor, overwrite with space, then \b again
                sys.stdout.write("\b \b")
                sys.stdout.flush()
                enteredPassword = enteredPassword[:-1]
        elif 0 <= key <= 31:
            # Do nothing for unprintable characters.
            # We ignore arrow keys, home, end, etc.
            pass
        else:
            # Key is part of the password; display the mask character.
            char = chr(key)
            sys.stdout.write(mask)
            sys.stdout.flush()
            enteredPassword.append(char)


def win_term_getchar(strip_escapes: bool = True) -> str:
    """
    Call msvcrt.getch(), optionally skipping escape sequences.
    """
    while True:
        ch = getch()

        # https://docs.python.org/3/library/msvcrt.html
        # If the pressed key was a special function key, this will return
        # '\000' or '\xe0'; the next call will return the keycode.

        if not strip_escapes:
            return ch

        if ch == "\x00" or ch == "\xe0":
            # discard next keycode, part of escape sequence
            getch()
        else:
            return ch


def unix_term_getchar(strip_escapes: bool = True) -> str:
    """
    Read the next character from stdin, optionally skipping past any ANSI
    escape sequences.
    """
    ch: str = getch()

    while True:
        # handle ^C
        if ch == "\x03":
            raise KeyboardInterrupt

        if not strip_escapes:
            return ch

        # All C0 control codes except for ^C we return as is

        # When strip_escapes is set, we remove ANSI escape sequences starting
        # with ^[ (ESC)

        if ch != "\x1B":
            # not ESC
            return ch

        # read next char after ESC
        ch = getch()

        # Determine whether the next char is part of the escape sequence.
        if ch == "[":
            # This is a multi-byte CSI (Control Sequence Introducer)
            # sequence

            # Read next char
            ch = getch()

            # consume any parameter bytes in 0x30-0x3F [0-?]
            while "\x30" <= ch <= "\x3F":
                ch = getch()

            # consume any intermediate bytes in 0x20-0x2F [ -/]
            while "\x20" <= ch <= "\x2F":
                ch = getch()

            if "\x40" <= ch <= "\x7E":
                # correct final byte, continue with next char
                ch = getch()
                continue
            else:
                # invalid final byte... invalid escape sequence
                raise ValueError("Invalid ANSI CSI escape sequence")

        elif "\x40" <= ch <= "\x5F":
            # This is a C1 Fe escape sequence of two bytes, discard char
            # and continue after next char
            ch = getch()
            continue

        else:
            # Char following ESC is not part of an escape sequence, retry
            # loop and process as standalone character
            continue

        raise NotImplementedError("notreached")


def termios_getch():
    # type: () -> str
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch


if sys.platform == "win32":
    # Windows
    import msvcrt

    getch = msvcrt.getch
    term_getchar = win_term_getchar

else:
    # macOS and Linux
    import termios
    import tty

    getch = termios_getch
    term_getchar = unix_term_getchar



import dataclasses
import secrets
import string
import textwrap


from .errors import InputError


@dataclasses.dataclass
class Alphabet:
    name: str
    chars: str
    passthrough: set[str]
    chars_dict: dict[str, int] = dataclasses.field(init=False)
    description: str = ""
    description_passthrough: str = ""
    aliases: set[str] = dataclasses.field(default_factory=set)
    decimal: bool = False
    specials_map: dict[str, str] | None = None

    def __post_init__(self) -> None:
        self.chars_dict = {v: i for i, v in enumerate(self.chars)}
        self._passthrough_trans = str.maketrans({c: None for c in self.passthrough})

    def prepare_key(self, text: str) -> str:
        """
        Given key text, prepare for use in encryption/decryption.
        - Remove passthrough characters.
        - If this alphabet has decimal=True, convert decimal input to chars.
        """
        cleaned = self.remove_passthrough(text)

        if self.decimal:
            return self.decimal_decode(cleaned)
        else:
            return cleaned

    def remove_passthrough(self, text: str) -> str:
        """
        Return the provided text with passthrough characters removed.
        """
        return text.translate(self._passthrough_trans)

    def generate_key(self, length: int, auto_decimal: bool = True) -> str:
        """
        Generate a key from this alphabet, using the `secrets` module CSPRNG.
        """
        key = "".join(secrets.choice(self.chars) for i in range(length))

        if self.decimal and auto_decimal:
            return self.decimal_encode(key)

        return key

    @property
    def aliases_str(self) -> str:
        return "|".join(sorted(self.aliases))

    @property
    def ansi_spaces(self) -> bool:
        """
        Default setting for whether output should use fancy ansi inverted color
        spaces. This will default to true if space is in the alphabet and this
        isn't a decimal alphabet.
        """
        return " " in self.chars_dict and not self.decimal

    @property
    def chars_for_display(self) -> str:
        specials_map = {"\0": "␀", "\t": "␉", "\n": "␊", "\f": "␌", "\r": "␍"}
        trans = str.maketrans(specials_map)
        return self.chars.translate(trans)

    @property
    def chars_escaped(self) -> list[str]:
        specials_map = {"\0": "\\0", "\t": "\\t", "\n": "\\n", "\f": "\\f", "\r": "\\r"}
        return [specials_map[c] if c in specials_map else c for c in self.chars]

    def _char_to_decimal(self, char: str) -> str:
        """
        Given a char, return a 2-digit int (base 10) index in the alphabet.

        e.g. if alphabet='decimal':
            "~" -> "99"
        """
        try:
            return "%02d" % self.chars_dict[char]
        except KeyError as err:
            raise InputError(
                f"Invalid input char {char!r} for alphabet {self.name!r}"
            ) from err

    def decimal_encode(self, text: str, wrap: int | None = 60) -> str:
        """
        Format string as a series of 2-digit int (base 10) indexes in the
        alphabet. Each 2-digit number is separated by spaces.

        e.g. if alphabet='decimal':
            "FOO!" -> "43 52 52 06"
        """

        encoded = " ".join(self._char_to_decimal(c) for c in text)
        if wrap:
            return textwrap.fill(encoded, width=wrap)
        return encoded

    def decimal_decode(self, decimals: str) -> str:
        """
        Given a string of whitespace separated decimals that represent indexes
        in the alphabet, convert to a string of chars in the alphabet.

        e.g. if alphabet='decimal':
            "43 52 52 06" -> "FOO!"
        """

        s = ""
        for dec in decimals.split():
            try:
                num = int(dec)
            except ValueError as err:
                raise InputError(f"Invalid decimal input: {dec!r}") from err

            if num < 0:
                raise InputError(f"Negative numbers are invalid: {num!r}")
            if num >= len(self.chars):
                maxnum = len(self.chars) - 1
                raise InputError(
                    f"Invalid input for alphabet: {dec!r} not in 0..{maxnum!r}"
                )

            s += self.chars[num]

        return s


ALPHABET_DECIMAL = Alphabet(
    name="decimal",
    decimal=True,
    chars="\0\t\n\f\r !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~",  # noqa: E501
    passthrough=set(),  # NB: \v not in passthrough, unlike printable
    description="100-char full ASCII, ciphertext written as digits",
    description_passthrough="none",
)

ALPHABET_PRINTABLE = Alphabet(
    name="printable",
    chars=" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~",  # noqa: E501
    passthrough={"\t", "\n", "\v", "\f", "\r"},
    description="All printable characters and spaces",
    description_passthrough="other whitespace",
)

ALPHABET_LETTERS_ONLY = Alphabet(
    name="letters",
    chars="ABCDEFGHIJKLMNOPQRSTUVWXYZ",
    passthrough=set(string.punctuation + string.whitespace),
    description="Uppercase letters only",
    description_passthrough="punctuation/whitespace",
)
ALPHABET_ALPHANUMERIC_UPPER = Alphabet(
    name="alpha-upper",
    chars="ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789",
    passthrough=set(string.punctuation + string.whitespace),
    description="Uppercase letters and numbers",
    description_passthrough="punctuation/whitespace",
)
ALPHABET_ALPHANUMERIC_MIXED = Alphabet(
    name="alpha-mixed",
    chars="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
    passthrough=set(string.punctuation + string.whitespace),
    description="Mixed case letters and numbers",
    description_passthrough="punctuation/whitespace",
)


ALPHABETS: dict[str, Alphabet] = {
    "decimal": ALPHABET_DECIMAL,
    "printable": ALPHABET_PRINTABLE,
    "letters": ALPHABET_LETTERS_ONLY,
    "alpha-mixed": ALPHABET_ALPHANUMERIC_MIXED,
    "alpha-upper": ALPHABET_ALPHANUMERIC_UPPER,
}


ALPHABET_ALIASES: dict[str, str] = {
    "100": "decimal",
    "ascii": "decimal",
    "print": "printable",
    "wheel": "printable",
    "upper": "letters",
    "uppercase": "letters",
    "alpha": "alpha-mixed",
    "alphanumeric": "alpha-mixed",
    "alphanumeric-upper": "alpha-upper",
    "alphanumeric-mixed": "alpha-mixed",
}


for alias, target in ALPHABET_ALIASES.items():
    ALPHABETS[target].aliases.add(alias)


def get_alphabet(name: str) -> Alphabet:
    """
    Look up an Alphabet by name or alias.
    """
    if not isinstance(name, str):
        raise ValueError(f"name must be str, got {name!r}")

    if name in ALPHABET_ALIASES:
        name = ALPHABET_ALIASES[name]

    return ALPHABETS[name]


def list_alphabets_labels(aliases: bool = True) -> str:
    """
    Print help text describing each alphabet.
    """
    return "\n".join(
        "\n".join(
            [
                "  " + a.name + ":",
                "      " + a.description,
                "      aliases: " + "(" + a.aliases_str + ")",
                "      passthrough: " + a.description_passthrough,
                "      chars: " + a.chars_for_display,
                "",
            ]
        )
        for a in ALPHABETS.values()
    )



import itertools
import operator
from typing import Callable, Iterator, Optional, TextIO

from .alphabet import Alphabet, get_alphabet
from .errors import CipherError, InputError
from .pwinput import pwinput


class Cipher:
    def __init__(
        self,
        key: Optional[str] = None,
        key_file: Optional[TextIO] = None,
        batch: bool = False,
        alphabet: Optional[Alphabet] = None,
        alphabet_name: Optional[str] = None,
        insecure_allow_broken_short_key: bool = False,
        max_key_size: int = 1024 * 1024,
        wrap_decimal_width: int | None = 60,
    ):
        """
        If the key is random and longer than the plaintext, then this is
        effectively a one-time pad (secure).

        If insecure_allow_broken_short_key is set to True, then we allow keys
        shorter than the plaintext to be repeated, which acts like the
        classical (insecure) Vigenere cipher, which can be easily broken by
        frequency analysis.
        """

        if key_file and key:
            raise InputError("Cannot pass both key and key_file")

        if key_file:
            key = key_file.read(max_key_size + 1)
            if len(key) > max_key_size:
                raise InputError(f"Exceeded max key size: {max_key_size} B")
        elif key is None:
            if batch:
                raise InputError("Must provide key")
            else:
                key = pwinput("Key: ")

        if not key:
            raise InputError("Empty key")

        # Set alphabet directly or by name
        if alphabet:
            if alphabet_name:
                raise InputError("Cannot pass both alphabet and alphabet_name")

            if not isinstance(alphabet, Alphabet):
                raise InputError(f"Expected Alphabet instance, got {alphabet!r}")

            self.alphabet = alphabet

        else:
            if not alphabet_name:
                raise InputError("Must pass alphabet or alphabet_name")

            self.alphabet = get_alphabet(name=alphabet_name)

        # remove passthrough chars and possibly convert from decimal
        self.key = self.alphabet.prepare_key(key)

        self.insecure_allow_broken_short_key = insecure_allow_broken_short_key

        self.wrap_decimal_width = wrap_decimal_width

    def encrypt(self, text: str) -> str:
        """Encrypt provided text."""
        output = self._crypt(text=text, op=operator.add, input_label="plaintext")

        # convert output to decimal as needed
        if self.alphabet.decimal:
            output = (
                self.alphabet.decimal_encode(output, wrap=self.wrap_decimal_width)
            ) + "\n"

        return output

    def decrypt(self, text: str) -> str:
        """Decrypt provided text."""

        # convert decimal input as needed
        if self.alphabet.decimal:
            text = self.alphabet.decimal_decode(text)

        return self._crypt(text=text, op=operator.sub, input_label="ciphertext")

    def _crypt(
        self,
        text: str,
        op: Callable[[int, int], int],
        input_label: str,
    ) -> str:
        """
        Generic function handling encrypt and decrypt
        """
        if text is None:
            raise InputError("Must provide text")

        if not self.insecure_allow_broken_short_key:
            if len(self.key) < len(self.alphabet.remove_passthrough(text)):
                raise CipherError(f"Key is shorter than {input_label}")

        output = ""

        iter_in = iter(text)
        iter_key: Iterator[str]

        if self.insecure_allow_broken_short_key:
            # loop key forever if insecure mode is enabled
            # This behaves like a classical Vigenere cipher (easily broken)
            # instead of like a one-time pad (secure).
            iter_key = itertools.cycle(self.key)
        else:
            # iterate once through the key
            iter_key = iter(self.key)

        while True:
            try:
                c = next(iter_in)
            except StopIteration:
                return output

            # pass through certain text without consuming key
            while c in self.alphabet.passthrough:
                output += c
                try:
                    c = next(iter_in)
                except StopIteration:
                    return output

            try:
                k = next(iter_key)
            except StopIteration:
                raise CipherError(
                    f"Unexpected (bug?) key is shorter than {input_label}"
                )

            try:
                c_int = self.alphabet.chars_dict[c]
            except KeyError:
                raise CipherError(
                    f"Invalid character for alphabet {self.alphabet.name!r}"
                    + f" in {input_label} input: {c!r}"
                )

            try:
                k_int = self.alphabet.chars_dict[k]
            except KeyError:
                raise CipherError(
                    f"Invalid character for alphabet {self.alphabet.name!r}"
                    + f" in key: {k!r}"
                )

            o_int = op(c_int, k_int) % len(self.alphabet.chars)
            o_chr = self.alphabet.chars[o_int]

            output += o_chr

        return output



import csv
import functools
import sys
from typing import Callable, Optional, ParamSpec, TextIO, TypeVar

import click
import strictyaml

from .alphabet import (
    ALPHABETS,
    Alphabet,
    get_alphabet,
    list_alphabets_labels,
)
from .cipher import Cipher
from .errors import CLIError

# make help available at -h as well as default --help
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


ALIASES = {
    "alpha": "alphabet",
    "alphabets": "alphabet",
    "d": "dec",
    "decrypt": "dec",
    "digits": "decimal",
    "e": "enc",
    "encrypt": "enc",
    "genkey": "keygen",
}


class AliasedGroup(click.Group):
    # @typing.override  # python 3.12+
    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        if cmd_name in ALIASES:
            cmd_name = ALIASES[cmd_name]
        return super().get_command(ctx, cmd_name)


@click.group(cls=AliasedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(package_name="vigenere-py")
def cli() -> None:
    """
    Vigenère cipher encryption for Python.

    The cipher alphabet of possible characters must be set by -a/--alphabet or
    by env var VIGENERE_ALPHABET. (See `vigenere alphabet` for list.)

    Run `vigenere COMMAND --help` for more info on each command.
    """


def validate_alphabet(
    ctx: click.Context, param: click.core.Parameter, value: str
) -> Alphabet:
    """
    Click param validator for Alphabet. Accepts string as input and returns an
    Alphabet. Prints error/usage message if no alphabet is found with that
    name.
    """

    if value == ALPHABET_UNSET:
        # click.echo(ctx.get_help(), err=True)
        click.echo(
            "\n".join(
                [
                    "Error: Must set option -a/--alphabet or env var VIGENERE_ALPHABET",
                    "\nKnown alphabets: " + ", ".join(ALPHABETS.keys()),
                    "(See `vigenere alphabets` for more info)",
                ]
            ),
            err=True,
        )
        ctx.exit(1)

    try:
        return get_alphabet(name=value)
    except KeyError:
        click.echo("Known alphabets:\n" + list_alphabets_labels(aliases=True), err=True)
        click.secho(
            "Error: Alphabet not found: " + repr(value), fg="red", bold=True, err=True
        )
        ctx.exit(1)


def validate_alphabet_optional(
    ctx: click.Context, param: click.core.Parameter, value: Optional[str]
) -> Optional[Alphabet]:
    """
    Click param validator for Alphabet. Accepts optional string as input and
    returns either None or an Alphabet.
    """

    if not value:
        return None

    return validate_alphabet(ctx=ctx, param=param, value=value)


ALPHABET_UNSET = "<unset>"


# Alphabet option is used by several commands
_alphabet_option = click.option(
    "-a",
    "--alphabet",
    help="Cipher alphabet, if not set by VIGENERE_ALPHABET",
    metavar="ALPHABET",
    default=ALPHABET_UNSET,
    envvar="VIGENERE_ALPHABET",
    callback=validate_alphabet,
)


P = ParamSpec("P")
R = TypeVar("R")


def _cipher_options(f: Callable[P, R]) -> Callable[P, R]:
    """
    Common options for encrypt and decrypt
    """

    @click.argument("input", type=click.File("r"), required=False)
    @_alphabet_option
    @click.option(
        "-b", "--batch", help="Non-interactive mode", is_flag=True, default=False
    )
    @click.option("-k", "--key-file", help="Key file", type=click.File("r"))
    @click.option("-o", "--output", help="Output file", type=click.File("w"))
    @click.option(
        "--insecure",
        is_flag=True,
        default=False,
        help="Allow short keys to loop (easily cracked!!)",
    )
    @functools.wraps(f)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
        return f(*args, **kwargs)

    return wrapped


@cli.command(name="enc")
@_cipher_options
def encrypt(
    input: Optional[TextIO],
    key_file: Optional[TextIO],
    output: Optional[TextIO],
    alphabet: Alphabet,
    batch: bool,
    insecure: bool,
) -> None:
    """
    Encrypt text with a Vigenère cipher.

    Read plaintext from INPUT file or from stdin if not provided.

    Prompt for key interactively if key file not given.

    For example:

        vigenere enc -o out.txt input.txt

    """

    if not input:
        input = sys.stdin

    try:
        c = Cipher(
            key_file=key_file,
            batch=batch,
            alphabet=alphabet,
            insecure_allow_broken_short_key=insecure,
        )
    except CLIError as err:
        click.secho("Error: " + str(err), fg="red")
        sys.exit(3)

    # If output is a TTY, highlight spaces in ANSI inverted colors, if we're
    # using an alphabet that may contain spaces.
    if output:
        ansi_invert_spaces = False
    else:
        ansi_invert_spaces = sys.stdout.isatty() and c.alphabet.ansi_spaces

    if input.isatty():
        click.echo("Text to encrypt:", err=True)

    try:
        ciphertext = c.encrypt(input.read())
    except CLIError as err:
        click.secho("Error: " + str(err), fg="red")
        sys.exit(3)

    if output:
        output.write(ciphertext)
    else:
        if input.isatty():
            click.echo("Ciphertext:", err=True)

        if ansi_invert_spaces:
            ciphertext = ciphertext.replace(" ", "\033[7m \033[27m")

        click.echo(ciphertext, nl=False)


@cli.command(name="dec")
@_cipher_options
def decrypt(
    input: Optional[TextIO],
    key_file: Optional[TextIO],
    output: Optional[TextIO],
    alphabet: Alphabet,
    batch: bool,
    insecure: bool,
) -> None:
    """
    Decrypt Vigenère ciphertext.

    Read plaintext from INPUT file or from stdin if not provided.

    Prompt for key interactively if key file not given.

    For example:

        vigenere dec cipher.txt
    """

    if not input:
        input = sys.stdin

    try:
        c = Cipher(
            key_file=key_file,
            batch=batch,
            alphabet=alphabet,
            insecure_allow_broken_short_key=insecure,
        )
    except CLIError as err:
        click.secho("Error: " + str(err), fg="red")
        sys.exit(3)

    if input.isatty():
        click.echo("Enter ciphertext...", err=True)

    try:
        plaintext = c.decrypt(input.read())
    except CLIError as err:
        click.secho("Error: " + str(err), fg="red")
        sys.exit(3)

    if output:
        output.write(plaintext)
    else:
        if input.isatty():
            click.echo("Plaintext:", err=True)
        click.echo(plaintext, nl=False)


@cli.command()
@click.argument("length", type=int)
@_alphabet_option
@click.option("-o", "--output", help="Write key to given file", type=click.File("w"))
@click.option(
    "-f",
    "--format",
    help="Output format",
    default="plain",
    type=click.Choice(["plain", "yaml"]),
)
def keygen(
    length: int,
    output: Optional[TextIO],
    alphabet: Alphabet,
    format: str,
) -> None:
    """
    Generate a random key, suitable for use as a one time pad.
    """

    alpha = alphabet
    key = alpha.generate_key(length=length)

    if format == "yaml":
        key = strictyaml.as_document({"key": key}).as_yaml()
    elif format == "plain":
        pass
    else:
        raise ValueError("Invalid format: " + repr(format))

    if output:
        output.write(key)
    else:
        ansi_invert_spaces = (
            sys.stdout.isatty() and format == "plain" and alpha.ansi_spaces
        )
        if ansi_invert_spaces:
            key = key.replace(" ", "\033[7m \033[27m")

        click.echo(key, nl=(format == "plain"))


@cli.command()
@click.argument("alphabet", required=False, callback=validate_alphabet_optional)
@click.option(
    "-f",
    "--format",
    help="Output format",
    default="plain",
    type=click.Choice(["plain", "tab", "csv"]),
)
@click.option("--tab", is_flag=True, help="Tab delimit output")
@click.option("--csv", "csv_out", is_flag=True, help="CSV format output")
@click.option("--table", is_flag=True, help="Print decimal table")
def alphabet(
    alphabet: Optional[Alphabet],
    format: str,
    label: Optional[str] = None,
    csv_out: bool = False,
    tab: bool = False,
    table: bool = False,
) -> None:
    """
    Print helpful info about supported alphabets.

    If a single alphabet is given, print the characters in that alphabet.

    Or, if no label is given, print details about all known alphabets.
    """

    if csv_out:
        format = "csv"
    if tab:
        format = "tab"

    if not alphabet:
        if format == "csv":
            writer = csv.writer(sys.stdout)
            header = ["name", "description", "aliases"]
            writer.writerow(header)

            for alpha in ALPHABETS.values():
                row = [alpha.name, alpha.description, alpha.aliases_str]
                writer.writerow(row)

        elif format == "tab":
            for alpha in ALPHABETS.values():
                row = [alpha.name, alpha.description, alpha.aliases_str]
                click.echo("\t".join(row))

        elif format == "plain":
            click.echo("Known alphabets:\n" + list_alphabets_labels(aliases=True))
        else:
            raise ValueError("Invalid format: " + repr(format))

        return

    alpha = alphabet

    if table:
        for i, c in enumerate(alpha.chars_escaped):
            click.echo(("%02d" % i) + "\t" + c)
        return

    if format == "csv":
        row = list(alpha.chars_escaped)
        writer = csv.writer(sys.stdout)
        writer.writerow(row)

    elif format == "tab":
        click.echo("\t".join(alpha.chars_escaped))

    elif format == "plain":
        click.echo(alpha.chars_for_display)

    else:
        raise ValueError(f"Bad format: {format!r}")


@cli.command()
@click.argument("file", type=click.File("r"), required=False)
@_alphabet_option
@click.option(
    "-e",
    "--encode",
    "mode",
    flag_value="encode",
    help="Encode string to decimal",
)
@click.option(
    "-d",
    "--decode",
    "mode",
    flag_value="decode",
    help="Decode decimal to string",
)
@click.option(
    "-w", "--wrap", metavar="WIDTH", default=60, help="Wrap decimal output at WIDTH"
)
@click.pass_context
def decimal(
    ctx: click.Context,
    alphabet: Alphabet,
    mode: str,
    file: Optional[TextIO],
    wrap: int,
) -> None:
    """
    Decimal encode/decode data and print to standard output.

    In -e/--encode mode, accept string as input and convert to digits according
    to each character's index in the cipher alphabet.

    In -d/--decode mode, accept space-separated digits as input and convert to
    a string of characters in the alphabet.

    For example:

        fortune | vigenere decimal -a 100 -e

        echo 45 74 81 81 84 | vigenere decimal -a 100 -d
    """

    if not mode:
        click.echo(ctx.get_help(), err=True)
        click.echo("\nError: Must set mode -e or -d", err=True)
        ctx.exit(1)

    if not file:
        if sys.stdin.isatty():
            click.echo("Enter input on stdin, end with EOF/ctrl^d...", err=True)

        file = sys.stdin

    text = file.read()

    try:
        if mode == "encode":
            text = alphabet.remove_passthrough(text)
            click.echo(alphabet.decimal_encode(text=text, wrap=wrap))
        elif mode == "decode":
            click.echo(
                alphabet.decimal_decode(decimals=text),
                nl="\n" not in alphabet.chars_dict,
            )
        else:
            raise ValueError(f"Invalid mode: {mode!r}")

    except CLIError as err:
        click.secho("Error: " + str(err), fg="red", bold=True)
        sys.exit(3)
